from itertools import chain, repeat, zip_longest
from typing import Iterable, Iterator, Sequence, Tuple

from rxn.utilities.misc import get_multipliers

from .miscellaneous import merge_reactions
from .reaction_equation import ReactionEquation, canonicalize_compounds, sort_compounds
from .reaction_smiles import (
    ReactionFormat,
    parse_any_reaction_smiles,
    to_reaction_smiles,
)
from .tokenization import detokenize_smiles


class ReactionCombiner:
    """
    Class to combine sets of precursors with sets of products, or sets of partial
    reactions with other sets of partial reactions.

    This class is typically useful when one needs to produce the full reaction
    SMILES starting from multiple files, such as A) one file for the precursors
    and one for the products, or B) two files containing each one part of a
    chemical equation.

    This class is particularly useful when the said files have different sizes,
    which can be the case when multiple predictions are made for each line of one
    of these files.
    """

    def __init__(
        self,
        standardize: bool = False,
        reaction_format: ReactionFormat = ReactionFormat.STANDARD_WITH_TILDE,
        fallback_reaction: str = ">>",
    ):
        """
        Args:
            standardize: whether to standardize (i.e. canonicalize and reorder) the reaction SMILES.
            reaction_format: which format should be used for the reaction SMILES.
            fallback_reaction: text / reaction to produce when a reaction is invalid.
        """
        self.standardize = standardize
        self.reaction_format = reaction_format
        self.fallback_reaction = fallback_reaction

    def combine(
        self, fragments_1: Sequence[str], fragments_2: Sequence[str]
    ) -> Iterator[str]:
        """
        See docstring of function ``combine_sequences``.
        """
        yield from self.combine_sequences(fragments_1, fragments_2)

    def combine_sequences(
        self, fragments_1: Sequence[str], fragments_2: Sequence[str]
    ) -> Iterator[str]:
        """
        Combine the two sequences of fragments into an iterator of reactions.

        Args:
            fragments_1: Sequence of sets of precursors strings (such as "CC.O.[Na+]~[Cl-]"),
                or list of partial reactions.
            fragments_2: Sequence of sets of product(s) strings, or list of partial
                reactions.

        Returns:
            Iterator over the resulting reaction SMILES.
        """
        fragments_1_multiplier, fragments_2_multiplier = self._get_multipliers(
            fragments_1, fragments_2
        )
        yield from self.combine_iterators(
            fragments_1=fragments_1,
            fragments_2=fragments_2,
            fragments_1_multiplier=fragments_1_multiplier,
            fragments_2_multiplier=fragments_2_multiplier,
        )

    def combine_iterators(
        self,
        fragments_1: Iterable[str],
        fragments_2: Iterable[str],
        fragments_1_multiplier: int = 1,
        fragments_2_multiplier: int = 1,
    ) -> Iterator[str]:
        """
        Combine the two iterators of fragments into an iterator of reactions.

        Args:
            fragments_1: Sequence of sets of precursors strings (such as "CC.O.[Na+]~[Cl-]"),
                or list of partial reactions.
            fragments_2: Sequence of sets of product(s) strings, or list of partial
                reactions.
            fragments_1_multiplier: how many times to duplicate the fragments_1.
            fragments_2_multiplier: how many times to duplicate the fragments_2.

        Raises:
            RuntimeError: if one of the iterators isn't fully consumed.
            ValueError: when one is not exactly a multiple of the other.

        Returns:
            Iterator over the resulting reaction SMILES.
        """

        self._validate_multipliers(fragments_1_multiplier, fragments_2_multiplier)

        # repeat itemwise the elements: https://stackoverflow.com/a/45799320
        fragment_1_iterator = chain.from_iterable(
            (repeat(e, fragments_1_multiplier) for e in fragments_1)
        )
        fragment_2_iterator = chain.from_iterable(
            (repeat(e, fragments_2_multiplier) for e in fragments_2)
        )

        for fragment_1, fragment_2 in zip_longest(
            fragment_1_iterator, fragment_2_iterator
        ):
            if fragment_1 is None or fragment_2 is None:
                raise RuntimeError("Mismatch in expected iterator length")
            yield self._to_reaction_smiles(fragment_1, fragment_2)

    def _to_reaction_smiles(self, fragment_1: str, fragment_2: str) -> str:
        try:
            return self._try_to_reaction_smiles(fragment_1, fragment_2)
        except Exception:
            return self.fallback_reaction

    def _try_to_reaction_smiles(self, fragment_1: str, fragment_2: str) -> str:
        # 1) get the initial reaction SMILES
        reaction_equation = self._to_raw_reaction(fragment_1, fragment_2)

        # 2) standardize if necessary
        if self.standardize:
            reaction_equation = sort_compounds(
                canonicalize_compounds(reaction_equation)
            )

        return to_reaction_smiles(
            reaction_equation, reaction_format=self.reaction_format
        )

    def _to_raw_reaction(self, fragment_1: str, fragment_2: str) -> ReactionEquation:
        """Get a ReactionEquation from the two strings."""
        fragment_1 = detokenize_smiles(fragment_1)
        fragment_2 = detokenize_smiles(fragment_2)

        fragment_1_is_reaction = ">" in fragment_1
        fragment_2_is_reaction = ">" in fragment_2

        # Case A: both are given in the reaction format
        if fragment_1_is_reaction and fragment_2_is_reaction:
            reaction_1 = parse_any_reaction_smiles(fragment_1)
            reaction_2 = parse_any_reaction_smiles(fragment_2)
            return merge_reactions(reaction_1, reaction_2)

        # Case A: fragment_1 represents the precursor(s), fragment_2 the product(s)
        if not fragment_1_is_reaction and not fragment_2_is_reaction:
            reaction_smiles = fragment_1 + ">>" + fragment_2
            return parse_any_reaction_smiles(reaction_smiles)

        raise ValueError(
            f'Cannot determine how to combine "{fragment_1}" and "{fragment_2}"'
        )

    def _get_multipliers(
        self, fragments_1: Sequence[str], fragments_2: Sequence[str]
    ) -> Tuple[int, int]:
        """Get the multipliers to use when iterating through the respective fragments.

        Raises:
            ValueError: when one is not exactly a multiple of the other.

        Returns:
            Tuple: fragments_1 multiplier, fragments_2 multiplier
        """
        a = len(fragments_1)
        b = len(fragments_2)

        m_a, m_b = get_multipliers(a, b)

        return m_a, m_b

    def _validate_multipliers(self, multiplier_1: int, multiplier_2: int) -> None:
        """
        Make sure that the given multipliers can be used with the reaction combiner.

        Raises:
            ValueError: when one is not exactly a multiple of the other.
        """
        # Fail if one is not exactly a multiple of the other
        if 1 not in {multiplier_1, multiplier_2}:
            raise ValueError(
                "The number of fragments of reactions are not an exact multiple of "
                f"each other: the multipliers are {multiplier_1} and {multiplier_2}."
            )



import pytest

from rxn.chemutils.exceptions import InvalidSmiles
from rxn.chemutils.smiles_standardization import (
    standardize_molecules,
    standardize_smiles,
)


def test_standardize_smiles() -> None:
    smiles = "C(O)C"
    # case 1: canonicalization and sanitization
    assert standardize_smiles(smiles) == "CCO"
    # case 2: disabled canonicalization
    assert standardize_smiles(smiles, canonicalize=False) == "C(O)C"
    # case 3: disabled canonicalization with inchification (inherent canonicalization)
    assert standardize_smiles(smiles, canonicalize=False, inchify=True) == "CCO"
    # case 4: canonicalization different from inchification (tautomers interconversion)
    smiles = "CNC(=O)C"
    assert standardize_smiles(smiles) == "CNC(C)=O"
    assert standardize_smiles(smiles, inchify=True) == "CN=C(C)O"
    # case 5: inchification with metal disconnection
    smiles = "CCCC[Li]"
    assert standardize_smiles(smiles, inchify=True) == "[CH2]CCC.[Li]"
    # case 6: testing an invalid SMILES and error handling
    invalid_smiles = "C%5%%5"
    with pytest.raises(InvalidSmiles):
        standardize_smiles(invalid_smiles)


def test_standardize_molecules() -> None:
    # successful cases with different standardization flavours
    # case 1: default fragment bond
    molecules = "C(O)C.CCO.CC~C"
    assert standardize_molecules(molecules) == "CCO.CCO.C~CC"
    # case 2: custom fragment bond
    molecules = "C(O)C.CCO.CC|C"
    assert standardize_molecules(molecules, fragment_bond="|") == "CCO.CCO.C|CC"
    # case 3: molecule token delimiter
    molecules = "C(O)C.CCO.CC~C._C_"
    assert (
        standardize_molecules(molecules, molecule_token_delimiter="_")
        == "C.CCO.CCO.C~CC"
    )
    # case 4: molecule token delimiter with disabled ordering (canonicalization order the fragments)
    molecules = "C(O)C.CCO.CC~C._C_"
    assert (
        standardize_molecules(
            molecules, molecule_token_delimiter="_", ordered_precursors=False
        )
        == "CCO.CCO.C~CC.C"
    )
    # case 5: molecule token delimiter with disabled ordering and canonicalization
    molecules = "C(O)C.CCO.CC~C._C_"
    assert (
        standardize_molecules(
            molecules,
            canonicalize=False,
            molecule_token_delimiter="_",
            ordered_precursors=False,
        )
        == "C(O)C.CCO.CC~C.C"
    )
    # case 6: molecule token delimiter with disabled ordering and canonicalization, but enabled inchification
    molecules = "C(O)C.CCO.CC~C._C_"
    assert (
        standardize_molecules(
            molecules,
            canonicalize=False,
            inchify=True,
            molecule_token_delimiter="_",
            ordered_precursors=False,
        )
        == "CCO.CCO.C~CC.C"
    )
    # expected failures due to mismatch between the molecules string and the standard
    # case 7: unexpected fragment bond
    molecules = "C(O)C.CCO.CC|C"
    with pytest.raises(InvalidSmiles):
        standardize_molecules(molecules)
    # case 8: unexpected molecule token delimiter
    molecules = "C(O)C.CCO.CC~C._C_"
    with pytest.raises(InvalidSmiles):
        standardize_molecules(molecules)



import random

import pytest
from rxn.utilities.basic import identity
from rxn.utilities.containers import all_identical

from rxn.chemutils.exceptions import InvalidSmiles
from rxn.chemutils.smiles_augmenter import SmilesAugmenter
from rxn.chemutils.smiles_randomization import randomize_smiles_rotated

single_compound = "O=C(C)Oc1ccccc1C(=O)O"
salt_compound = "CC[NH+](CC)CC.CC(=O)[O-]"
multismiles = "O=C(C)Oc1ccccc1C(=O)O.Cl.[Na+]~[Cl-]"
rxn_smiles_1 = "CC(=O)Cl.NCC.[Na+]~[Cl-]>>CC(=O)NCC"
rxn_smiles_2 = "CC(=O)Cl.NCC.[Na+].[Cl-]>>CC(=O)NCC |f:2.3|"  # extended format
rxn_smiles_3 = "CC(=O)Cl.NCC.[Na+]~[Cl-]>>CC(=O)NCC.Cl"  # multiple products


def test_do_nothing() -> None:
    # If there is no SMILES randomization and no shuffling, the class should
    # just return the same strings back
    augmenter = SmilesAugmenter(augmentation_fn=identity, shuffle=False)

    test_strings = [
        single_compound,
        salt_compound,
        multismiles,
        rxn_smiles_1,
        rxn_smiles_2,
        rxn_smiles_3,
    ]
    for test_smiles in test_strings:
        assert augmenter.augment(test_smiles, 4) == 4 * [test_smiles]


def test_shuffling_only() -> None:
    augmenter = SmilesAugmenter(augmentation_fn=identity, shuffle=True)

    # On single compound: no change at all
    assert augmenter.augment(single_compound, 4) == 4 * [single_compound]

    # On salt compound: just reorders it -> 2 possible combinations
    assert len(set(augmenter.augment(salt_compound, 20))) == 2

    # On multismiles: fragments stay together; this leads to a total of 6 combinations
    assert len(set(augmenter.augment(multismiles, 60))) == 6

    # On reaction SMILES with two products: 6 combinations for precursors, 2 for products
    assert len(set(augmenter.augment(rxn_smiles_3, 150))) == 12


def test_smiles_augmentation_only() -> None:
    def dummy_augmentation(smiles: str) -> str:
        # Replace the SMILES strings by their lengths.
        return str(len(smiles))

    augmenter = SmilesAugmenter(augmentation_fn=dummy_augmentation, shuffle=False)

    query_and_expected = [
        (single_compound, "21"),
        (salt_compound, "13.10"),
        (multismiles, "21.2.11"),
        (rxn_smiles_1, "8.3.11>>9"),
        (rxn_smiles_2, "8.3.11>>9"),
        (rxn_smiles_3, "8.3.11>>9.2"),
    ]
    for query, expected in query_and_expected:
        # Note: no randomness, as the dummy augmentation is not random, and
        # there is no shuffling
        assert augmenter.augment(query, 4) == 4 * [expected]


def test_smiles_augmentation_only_with_probability() -> None:
    def dummy_augmentation(smiles: str) -> str:
        # Replace the SMILES strings by their lengths.
        return str(len(smiles))

    augmenter = SmilesAugmenter(
        augmentation_fn=dummy_augmentation, augmentation_probability=0.5, shuffle=False
    )

    # On single compound: either augmented or not
    assert set(augmenter.augment(single_compound, 10)) == {"21", single_compound}

    # To illustrate the behavior on strings with multiple SMILES, we just test it on
    # the multismiles input. There will be a mix of replaced and not replaced
    assert set(augmenter.augment(multismiles, 50)) == {
        "O=C(C)Oc1ccccc1C(=O)O.Cl.[Na+]~[Cl-]",
        "21.Cl.[Na+]~[Cl-]",
        "O=C(C)Oc1ccccc1C(=O)O.2.[Na+]~[Cl-]",
        "O=C(C)Oc1ccccc1C(=O)O.Cl.11",
        "21.2.[Na+]~[Cl-]",
        "21.Cl.11",
        "O=C(C)Oc1ccccc1C(=O)O.2.11",
        "21.2.11",
    }

    # No augmentation at all if the probability is zero
    augmenter = SmilesAugmenter(
        augmentation_fn=dummy_augmentation, augmentation_probability=0.0, shuffle=False
    )
    assert set(augmenter.augment(multismiles, 10)) == {multismiles}


def test_mix_augmentation_and_shuffling() -> None:
    augmenter = SmilesAugmenter(
        augmentation_fn=randomize_smiles_rotated,
        augmentation_probability=0.5,
        shuffle=True,
    )

    random.seed(42)
    assert augmenter.augment(rxn_smiles_3, 6) == [
        "NCC.CC(=O)Cl.[Cl-]~[Na+]>>Cl.CC(=O)NCC",
        "NCC.ClC(=O)C.[Na+]~[Cl-]>>Cl.N(CC)C(C)=O",
        "C(=O)(Cl)C.[Na+]~[Cl-].CCN>>CC(=O)NCC.Cl",
        "NCC.CC(=O)Cl.[Na+]~[Cl-]>>CC(=O)NCC.Cl",
        "NCC.[Na+]~[Cl-].CC(=O)Cl>>Cl.CC(=O)NCC",
        "CC(=O)Cl.[Na+]~[Cl-].C(C)N>>CC(=O)NCC.Cl",
    ]


def test_reproducibility() -> None:
    # illustrate the reproducibility at the example of a reaction SMILES
    augmenter = SmilesAugmenter(
        augmentation_fn=randomize_smiles_rotated,
        augmentation_probability=0.5,
        shuffle=True,
    )

    # When resetting the seed, we should always get exactly the same results
    results = []
    for _ in range(10):
        random.seed(42)
        results.append(augmenter.augment(rxn_smiles_3, 5))
    assert all_identical(results)

    # sampling one more time without resetting the seed -> results change
    results.append(augmenter.augment(rxn_smiles_3, 5))
    assert not all_identical(results)


def test_augmentation_errors() -> None:
    augmenter = SmilesAugmenter(
        augmentation_fn=randomize_smiles_rotated, ignore_exceptions=True
    )

    invalid_smiles = "thisisinvalid"

    # When errors are ignored: returns the original input
    assert augmenter.augment(invalid_smiles, 1) == [invalid_smiles]

    # When errors are not ignored: raises an exception
    augmenter.ignore_exceptions = False
    with pytest.raises(InvalidSmiles):
        _ = augmenter.augment(invalid_smiles, 1)



from collections import Counter
from typing import List

import pytest

from rxn.chemutils.exceptions import InvalidReactionSmiles, InvalidSmiles
from rxn.chemutils.miscellaneous import (
    apply_to_any_smiles,
    apply_to_smiles_groups,
    atom_type_counter,
    canonicalize_any,
    equivalent_smiles,
    get_individual_compounds,
    is_valid_smiles,
    merge_reactions,
    remove_chiral_centers,
    remove_double_bond_stereochemistry,
    smiles_has_atom_mapping,
    sort_any,
)
from rxn.chemutils.reaction_equation import ReactionEquation


def test_equivalent_smiles() -> None:
    assert equivalent_smiles("CCO", "C(C)O")
    assert not equivalent_smiles("CCO", "O(C)C")

    assert equivalent_smiles("CCO", "C(C)O", "OCC")
    assert not equivalent_smiles("CCO", "C(C)O", "O(C)C")
    assert not equivalent_smiles("CF=C", "F(C)C")

    # Should be false for invalid SMILES
    assert not equivalent_smiles("CCO", "", "O(C)C")

    # In case of invalid valence: not equivalent if check_valence flag is specified
    assert equivalent_smiles("CFC", "F(C)C")
    assert not equivalent_smiles("CFC", "F(C)C", check_valence=True)


def test_is_valid_smiles() -> None:
    assert is_valid_smiles("CCO")
    assert is_valid_smiles("C")
    assert is_valid_smiles("NOOOOO")
    assert is_valid_smiles("CC(CCC)")

    assert not is_valid_smiles("CFC")
    assert not is_valid_smiles("YES")
    assert not is_valid_smiles("Noo")
    assert not is_valid_smiles("CC(CC")

    assert is_valid_smiles("CFC", check_valence=False)
    assert not is_valid_smiles("YES", check_valence=False)
    assert not is_valid_smiles("CC(CC", check_valence=False)


def test_atom_type_counter() -> None:
    assert atom_type_counter("CCO") == Counter({"C": 2, "H": 6, "O": 1})
    assert atom_type_counter("CC=O") == Counter({"C": 2, "H": 4, "O": 1})
    assert atom_type_counter("[Na+].[Cl-]") == Counter({"Na": 1, "Cl": 1})

    # with radicals
    assert atom_type_counter("[C]CO") == Counter({"C": 2, "H": 3, "O": 1})

    with pytest.raises(InvalidSmiles):
        _ = atom_type_counter("CC(")


def test_remove_chiral_centers() -> None:
    input_expected_dict = {
        "O[C@](Br)(C)N": "O[C](Br)(C)N",
        "C[C@@](Br)(O)N": "C[C](Br)(O)N",
        "N[C@H](O)C": "N[CH](O)C",
        "N1[C@H](Cl)[C@@H](Cl)C(Cl)CC1": "N1[CH](Cl)[CH](Cl)C(Cl)CC1",
        "F/C=C/F": "F/C=C/F",
        "[N@+]": "[N+]",
        "[N@@+]": "[N+]",
        "[Si@@]": "[Si]",
        "[Si@H]": "[SiH]",
    }

    assert all(
        remove_chiral_centers(smi) == expected
        for smi, expected in input_expected_dict.items()
    )


def test_remove_double_bond_stereochemistry() -> None:
    input_expected_dict = {
        "C/C=C/C": "CC=CC",
        r"C/C=C\C": "CC=CC",
        r"C\C=C\C": "CC=CC",
        "F/C=C/C=C/C": "FC=CC=CC",
        "C1=CC=CC=C1": "C1=CC=CC=C1",
        "c1ccccc1": "c1ccccc1",
    }

    assert all(
        remove_double_bond_stereochemistry(smi) == expected
        for smi, expected in input_expected_dict.items()
    )


def test_apply_to_any_smiles() -> None:
    def dummy(smiles: str) -> str:
        return smiles + "0"

    # Single-component SMILES
    # Note: the first one is not considered to be a multi-component SMILES,
    # except with the "force_multicomponent" flag.
    assert apply_to_any_smiles("A.C.C.B", dummy) == "A.C.C.B0"
    assert (
        apply_to_any_smiles("A.C.C.B", dummy, force_multicomponent=True)
        == "A0.C0.C0.B0"
    )
    assert apply_to_any_smiles("CBA", dummy) == "CBA0"

    # Multi-component SMILES (note that here, the fragments are not reordered)
    assert apply_to_any_smiles("A.D~C.B", dummy) == "A0.D~C0.B0"

    # reaction SMILES
    assert apply_to_any_smiles("B.A.E~D.A>>C.B", dummy) == "B0.A0.E~D0.A0>>C0.B0"
    assert apply_to_any_smiles("A.E.D.A>>C |f:1.2|", dummy) == "A0.A0.E.D0>>C0 |f:2.3|"


def test_apply_to_smiles_groups() -> None:
    def dummy(smiles_list: List[str]) -> List[str]:
        return list(reversed(smiles_list))

    # Multi-component SMILES (note that here, the fragments are not reordered)
    # Note: Even if there is no "~", we consider the input to be a multicomponent SMILES
    assert apply_to_smiles_groups("A.D~C.B", dummy) == "B.D~C.A"
    assert apply_to_smiles_groups("A.C.C.B", dummy) == "B.C.C.A"
    assert apply_to_smiles_groups("CBA", dummy) == "CBA"

    # reaction SMILES
    assert apply_to_smiles_groups("B.A.E~D.A>>C.B", dummy) == "A.E~D.A.B>>B.C"
    assert apply_to_smiles_groups("A.B.E.D>>C |f:2.3|", dummy) == "E.D.B.A>>C |f:0.1|"


def test_canonicalize_any_on_molecule_smiles() -> None:
    assert canonicalize_any("CC(C)") == "CCC"
    assert canonicalize_any("[Na+].[Cl-]") == "[Cl-].[Na+]"
    assert canonicalize_any("CF(C)", check_valence=False) == "CFC"
    with pytest.raises(InvalidSmiles):
        _ = canonicalize_any("CF(C)")


def test_canonicalize_any_on_multicomponent_smiles() -> None:
    # Basic example
    assert canonicalize_any("CC(C).C(O)") == "CCC.CO"

    # Note: this will be assumed to be a single-compound SMILES -> reordered!
    assert canonicalize_any("C(O).CC(C)") == "CCC.CO"

    # Note: within the same compound: reordering due to compound canonicalization
    # Although "CCC" comes first alphabetically, it stays at the end.
    assert canonicalize_any("CO.C~O.CCC") == "CO.C~O.CCC"
    assert canonicalize_any("CO.O~C.CCC") == "CO.C~O.CCC"

    # Valence checking
    assert canonicalize_any("C(O)~CF(C)", check_valence=False) == "CFC~CO"
    with pytest.raises(InvalidSmiles):
        _ = canonicalize_any("CO.CFC")


def test_canonicalize_any_on_reaction_smiles() -> None:
    # Basic examples
    assert canonicalize_any("CC(C)>>C(O)") == "CCC>>CO"
    assert canonicalize_any("CC(C)>C(O)>C(O)") == "CCC>CO>CO"

    # Basic involving fragments
    assert canonicalize_any("CO.O~C>>C(O)") == "CO.C~O>>CO"
    assert canonicalize_any("CO.O.C>>C(O) |f:1.2|") == "CO.C.O>>CO |f:1.2|"
    # Note here the slight reordering - due to fragment parsing
    assert canonicalize_any("CC.CC.C.C>>CC |f:0.2|") == "CC.C.C.CC>>CC |f:2.3|"

    # Valence checking, and other potential exceptions
    assert canonicalize_any("CF(C)>>C(O)", check_valence=False) == "CFC>>CO"
    with pytest.raises(InvalidSmiles):
        _ = canonicalize_any("CFC>>CO")
    with pytest.raises(InvalidReactionSmiles):
        _ = canonicalize_any("CC>CO")
    with pytest.raises(InvalidReactionSmiles):
        _ = canonicalize_any("CC>>CC>>C(O)")


def test_canonicalize_any_with_fallback_value() -> None:
    fallback = "some_fallback_value"

    # No error -> fallback value not used
    assert canonicalize_any("C(C)C", fallback_value=fallback) == "CCC"
    assert canonicalize_any("CO>>C(C)C", fallback_value=fallback) == "CO>>CCC"

    # Error -> returns the fallback value
    assert canonicalize_any("CoMo", fallback_value=fallback) == fallback
    assert canonicalize_any("invalid>>C(C)C", fallback_value=fallback) == fallback


def test_canonicalize_any_with_sorting() -> None:
    # Note: the sorting in `N~OO` comes from the RDKit canonicalization of that compound.
    assert canonicalize_any("C(C).OO~N", sort_molecules=True) == "CC.N~OO"
    assert canonicalize_any("OO~N.C(C)", sort_molecules=True) == "CC.N~OO"
    assert canonicalize_any("OO~N.C(C)>>N.C", sort_molecules=True) == "CC.N~OO>>C.N"


def test_sort_any() -> None:
    # Single-component SMILES
    assert sort_any("A.C.C.B") == "A.B.C.C"
    assert sort_any("CBA") == "CBA"

    # Multi-component SMILES (note that here, the fragments are not reordered)
    assert sort_any("A.D~C.B") == "A.B.D~C"

    # reaction SMILES
    assert sort_any("B.A.E~D.A>>C.B") == "A.A.B.E~D>>B.C"
    assert sort_any("B.A.E.D.A>>C.B |f:2.3|") == "A.A.B.E.D>>B.C |f:3.4|"


def test_get_individual_compounds() -> None:
    # Single-component SMILES and/or multi-component SMILES
    assert get_individual_compounds("A.C.C.B") == ["A", "C", "C", "B"]
    assert get_individual_compounds("A.D~C.B") == ["A", "D.C", "B"]
    assert get_individual_compounds("CBA") == ["CBA"]

    # reaction SMILES
    assert get_individual_compounds("B.A.E~D.A>>C.B") == [
        "B",
        "A",
        "E.D",
        "A",
        "C",
        "B",
    ]
    assert get_individual_compounds("B.A.E.D.A>>C.B |f:2.3|") == [
        "B",
        "A",
        "A",
        "E.D",  # Note that it is located elsewhere, due to how extended SMILES are parsed
        "C",
        "B",
    ]


def test_merge_reactions() -> None:
    # No reaction given
    assert merge_reactions() == ReactionEquation([], [], [])

    # one reaction
    r1 = ReactionEquation(["A", "B"], ["C"], ["D"])
    assert merge_reactions(r1) == r1

    # two reactions
    r2 = ReactionEquation(["F"], ["G"], [])
    assert merge_reactions(r1, r2) == ReactionEquation(
        ["A", "B", "F"], ["C", "G"], ["D"]
    )

    # more than two reactions
    r3 = ReactionEquation([], ["H"], [])
    r4 = ReactionEquation([], [], ["I"])
    assert merge_reactions(r1, r2, r3, r4) == ReactionEquation(
        ["A", "B", "F"], ["C", "G", "H"], ["D", "I"]
    )


def test_smiles_has_atom_mapping() -> None:
    assert smiles_has_atom_mapping("[CH3:9][CH:8]([CH3:10])c1ccccc1")
    assert smiles_has_atom_mapping("CCO[CH3:9]")
    assert smiles_has_atom_mapping("C[CH3:9].O")

    assert not smiles_has_atom_mapping("CC")
    assert not smiles_has_atom_mapping("C[CH3]")
    assert not smiles_has_atom_mapping("C[12CH3]")

    # Note: does not care about valence, but SMILES must be valid
    # invalid valence -> call succeeds:
    assert smiles_has_atom_mapping("CF(C)CC[C:9]")
    # invalid SMILES -> call raises:
    with pytest.raises(InvalidSmiles):
        _ = canonicalize_any("[CH3:9]C(")



from typing import Any, Dict, Iterator, Optional, Tuple

import attr

from .property_parser import parse_properties


@attr.s(auto_attribs=True)
class Compound:
    """Wrapper for the compound dicts, which always contain a MOLSTRUCTURE,
    usually a SYMBOL, and sometimes a MOL_CAPTION."""

    mol_structure: str
    symbol: Optional[str]
    mol_caption: Optional[str]
    category: Optional[str]

    def get_name(self) -> Optional[str]:
        """Try to get a name by first looking at the SYMBOL and then at the MOLCAPTION."""
        if self.symbol is not None:
            return self.symbol
        return self.mol_caption

    def to_dict(self) -> Dict[str, str]:
        """Convert back to a dictionary, as in the RdfReaction properties."""
        d = {
            "MOLSTRUCTURE": self.mol_structure,
        }
        if self.symbol is not None:
            d["SYMBOL"] = self.symbol
        if self.mol_caption is not None:
            d["MOL_CAPTION"] = self.mol_caption
        return d

    @classmethod
    def from_dict(
        cls, compound_dict: Dict[str, str], category: Optional[str] = None
    ) -> "Compound":
        if category == "":
            category = None

        return cls(
            mol_structure=compound_dict["MOLSTRUCTURE"],
            symbol=compound_dict.get("SYMBOL", None),
            mol_caption=compound_dict.get("MOL_CAPTION", None),
            category=category,
        )


class ReactionProperties:
    """
    Class compiling the reaction properties given for an RDF reaction.
    """

    def __init__(self, meta: Dict[str, str]):
        pruned_meta = self._prune_unnecessary_lists(meta)
        parsed_properties = parse_properties(pruned_meta)
        try:
            self.properties: Dict[str, Any] = parsed_properties["RXN"]
        except KeyError:
            self.properties = parsed_properties

    def get_compound_dicts(self) -> Iterator[Dict[str, str]]:
        """NB: this function allows to iterate through the compounds in the
        property dict and make edits there (see ReactionUpdater)."""
        yield from find_compounds(self.properties)

    def get_compounds(self) -> Iterator[Compound]:
        """Get the compounds involved in a reaction, without the reactants and products."""
        yield from (
            Compound.from_dict(cpd_dict, category)
            for cpd_dict, category in find_compounds_with_category(self.properties)
        )

    def _prune_unnecessary_lists(self, meta: Dict[str, str]) -> Dict[str, str]:
        """
        In the meta dict, often there will be unnecessary list-like keys.
        In particular, "MOL(1):" and "SYMBOL(1):" can be removed as they are
        never necessary.
        """

        def simplify_key(key: str) -> str:
            updated_key = key.replace("MOL(1):", "").replace("SYMBOL(1):", "")
            if "MOL(" in updated_key or "SYMBOL(" in updated_key:
                raise ValueError(f"MOL or SYMBOL list with index other than 1: {key}")
            return updated_key

        return {simplify_key(key): value for key, value in meta.items()}


def find_compounds(property_dict: Dict[str, Any]) -> Iterator[Dict[str, str]]:
    """Iterate through a property dictionary to find compound dicts."""
    yield from (
        compound_dict
        for compound_dict, _ in find_compounds_with_category(property_dict)
    )


def find_compounds_with_category(
    property_dict: Dict[str, Any]
) -> Iterator[Tuple[Dict[str, str], str]]:
    """Iterate through a property dictionary to find compound dicts, with the
    associated category ("SOLVENT", "CATALYST", etc.)."""
    yield from _find_compounds_with_prefix("", property_dict)


def _find_compounds_with_prefix(
    prefix: str, property_dict: Dict[str, Any]
) -> Iterator[Tuple[Dict[str, str], str]]:
    if "MOLSTRUCTURE" in property_dict:
        yield property_dict, prefix

    for key, value in property_dict.items():
        if isinstance(value, dict):
            yield from _find_compounds_with_prefix("", value)

        # If the item is a list, we take the key as a prefix - this may be
        # something such as "SOLVENT", "CATALYST", etc.
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    yield from _find_compounds_with_prefix(key, item)
